# Copyright 2021-2023 Ian Jauslin
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

## elements that polyominoes are made of
import math
import sys

from point import Point,l_infinity,l_2
from tools import isint_nonzero,sgn,in_interval,ceil_grid,floor_grid
from kivy.graphics import Rectangle,Ellipse,Line

# parent class of all elements
class Element():
    def __init__(self,x,y,size,**kwargs):
        self.pos=Point(x,y)
        self.size=size

    # set position
    def setpos(self,x,y):
        self.pos.x=x
        self.pos.y=y


    # override in each subclass
    # draw element
    def draw(self,painter):
        return

    # override in each subclass
    # draw boundary
    def stroke(self,painter):
        return

    # override in each subclass
    # check whether an element interacts with square
    def check_interaction(self,element):
        return False

    # override in each subclass
    # whether x is in the support of the element
    def in_support(self,x):
        return False

    # override in each subclass
    # check whether an element is touching self
    def check_touch(self,element):
        return False

    # override in each subclass
    # find position along a line that comes in contact with the line going through element.pos in direction v
    def move_on_line_to_stick(self,element,v):
        return Point(0,0)

    # override in each subclass
    # move along edge of element
    # delta is the impossible move that was asked for
    def move_along(self,delta,element):
        return element


# rectangular element
# the size of the y component is specified by an aspect ratio: size_x=size, size_y=size*aspect
class Element_square(Element):
    def __init__(self,x,y,size,**kwargs):
        self.pos=Point(x,y)
        self.size=size
        self.aspect=kwargs.get("aspect",1.0)

    # draw element
    def draw(self,painter):
        Rectangle(pos=(painter.pos_tocoord_x(self.pos.x-0.5*self.size),painter.pos_tocoord_y(self.pos.y-0.5*self.size*self.aspect)),size=(self.size*painter.base_size,self.size*self.aspect*painter.base_size))

    # draw boundary
    def stroke(self,painter):
        # convert to graphical coordinates
        coordx=painter.pos_tocoord_x(square.pos.x)
        coordy=painter.pos_tocoord_y(square.pos.y)

        Line(points=(
        *(coordx-0.5*self.size*painter.base_size,coordy-0.5*self.size*self.aspect*painter.base_size),
        *(coordx-0.5*self.size*painter.base_size,coordy+0.5*self.size*self.aspect*painter.base_size),
        *(coordx+0.5*self.size*painter.base_size,coordy+0.5*self.size*self.aspect*painter.base_size),
        *(coordx+0.5*self.size*painter.base_size,coordy-0.5*self.size*self.aspect*painter.base_size),
        *(coordx-0.5*self.size*painter.base_size,coordy-0.5*self.size*self.aspect*painter.base_size)
        ))

    # for use with lattices
    # list of lattice points covered by square
    def lattice_points(self,lattice):
        out=[]
        dx=math.floor(0.5*self.size/lattice.spacing)
        dy=math.floor(0.5*self.size*self.aspect/lattice.spacing)
        for i in range(-dx,dx+1):
            for j in range(-dy,dy+1):
                out.append(Point(self.pos.x+i*lattice.spacing,self.pos.y+j*lattice.spacing))
        return out


    # check whether an element interacts with square
    # TODO: this only works if element is a rectangle!
    def check_interaction(self,element):
        # allow for error
        return max(abs(element.pos.x-self.pos.x)/(self.size+element.size),abs(element.pos.y-self.pos.y)/(self.size*self.aspect+element.size*element.aspect))<1/2-1e-11

    # whether x is in the support of the element
    def in_support(self,x):
        return max(abs(self.pos.x-x.x),abs(self.pos.y-x.y)/self.aspect)<=1/2

    # check whether an element is touching self
    # TODO: this only works if element is a rectangle!
    def check_touch(self,element):
        # allow for error
        if in_interval(max(abs(element.pos.x-self.pos.x)/(self.size+element.size),abs(element.pos.y-self.pos.y)/(self.size*self.aspect+element.size*element.aspect)),1/2-1e-11,1/2+1e-11):
                return True
        return False

    # find position along a line that comes in contact with the line going through element.pos in direction v
    # TODO: this only works if element is a rectangle!
    def move_on_line_to_stick(self,element,v):
        size_x=(self.size+element.size)/2
        size_y=(self.size*self.aspect+element.size*element.aspect)/2
        # compute intersections with four lines making up square
        if v.x!=0:
            if v.y!=0:
                intersections=[\
                Point(self.pos.x+size_x,element.pos.y+v.y/v.x*(self.pos.x+size_x-element.pos.x)),\
                Point(self.pos.x-size_x,element.pos.y+v.y/v.x*(self.pos.x-size_x-element.pos.x)),\
                Point(element.pos.x+v.x/v.y*(self.pos.y+size_y-element.pos.y),self.pos.y+size_y),\
                Point(element.pos.x+v.x/v.y*(self.pos.y-size_y-element.pos.y),self.pos.y-size_y)\
                ]
            else:
                intersections=[\
                Point(self.pos.x+size_x,element.pos.y),\
                Point(self.pos.x-size_x,element.pos.y)
                ]
        else:
            if v.y!=0:
                intersections=[\
                Point(element.pos.x,self.pos.y+size_y),\
                Point(element.pos.x,self.pos.y-size_y)\
                ]
            else:
                print("error: move_on_line_to_stick called with v=0, please file a bug report with the developer",file=sys.stderr)
                exit(-1)

        # compute closest one, on square
        closest=None
        dist=math.inf
        for i in range(0,len(intersections)):
            # check that it is on square
            if abs(intersections[i].x-self.pos.x)<=size_x+1e-11 and abs(intersections[i].y-self.pos.y)<=size_y+1e-11:
                if (intersections[i]-element.pos)**2<dist:
                    closest=intersections[i]
                    dist=(intersections[i]-element.pos)**2

        if closest==None:
            print("error: cannot move particle at (",element.pos.x,",",element.pos.y,") to the boundary of (",self.pos.x,",",self.pos.y,") in direction (",v.x,",",v.y,")",file=sys.stderr)
            exit(-1)

        # return difference to pos
        return closest-element.pos

    # move along edge of square
    # TODO: this only works if element is a rectangle!
    def move_along(self,delta,element):
        size_x=(self.size+element.size)/2
        size_y=(self.size*self.aspect+element.size*element.aspect)/2
        rel=element.pos-self.pos
        # check if the particle is stuck in the x direction
        if isint_nonzero(rel.x/size_x):
            # check y direction
            if isint_nonzero(rel.y/size_y):
                # in corner
                if sgn(delta.y)==-sgn(rel.y):
                    # stuck in x direction
                    return self.move_stuck_x(delta,element)
                elif sgn(delta.x)==-sgn(rel.x):
                    # stuck in y direction
                    return self.move_stuck_y(delta,element)
                # stuck in both directions
                return element.pos
            else:
                # stuck in x direction
                return self.move_stuck_x(delta,element)
        elif isint_nonzero(rel.y/size_y):
            # stuck in y direction
            return self.move_stuck_y(delta,element)
        # this should never happen
        else:
            print("error: stuck particle has non-integer relative position: (",rel.x,",",rel.y,")",file=sys.stderr)
            exit(-1)
    # move when stuck in the x direction
    def move_stuck_x(self,delta,element):
        size_y=(self.size*self.aspect+element.size*element.aspect)/2
        # only move in y direction
        candidate=Point(0,delta.y)
        # do not move past corners
        rel=element.pos.y-self.pos.y
        if delta.y>0:
            if rel<ceil_grid(rel,size_y)-1e-11 and delta.y+rel>ceil_grid(rel,size_y)+1e-11 and ceil_grid(rel,size_y)!=0:
                # stick to corner
                candidate.y=ceil_grid(rel,size_y)+self.pos.y-element.pos.y
        else:
            if rel>floor_grid(rel,size_y)+1e-11 and delta.y+rel<floor_grid(rel,size_y)-1e-11 and floor_grid(rel,size_y)!=0:
                # stick to corner
                candidate.y=floor_grid(rel,size_y)+self.pos.y-element.pos.y
        return candidate
    # move when stuck in the y direction
    def move_stuck_y(self,delta,element):
        size_x=(self.size+element.size)/2
        # onlx move in x direction
        candidate=Point(delta.x,0)
        # do not move past corners
        rel=element.pos.x-self.pos.x
        if delta.x>0:
            if rel<ceil_grid(rel,size_x)-1e-11 and delta.x+rel>ceil_grid(rel,size_x)+1e-11 and ceil_grid(rel,size_x)!=0:
                # stick to corner
                candidate.x=ceil_grid(rel,size_x)+self.pos.x-element.pos.x
        else:
            if rel>floor_grid(rel,size_x)+1e-11 and delta.x+rel<floor_grid(rel,size_x)-1e-11 and floor_grid(rel,size_x)!=0:
                # stick to corner
                candidate.x=floor_grid(rel,size_x)+self.pos.x-element.pos.x
        return candidate


# circular elements
# (size is the diameter)
class Element_circle(Element):
    # draw element
    def draw(self,painter):
        Ellipse(pos=(painter.pos_tocoord_x(self.pos.x-0.5*self.size),painter.pos_tocoord_y(self.pos.y-0.5*self.size)),size=(self.size*painter.base_size,self.size*painter.base_size))

    # draw boundary
    def stroke(self,painter):
        Line(circle=(painter.pos_tocoord_x(self.pos.x),painter.pos_tocoord_y(self.pos.y),self.size*0.5*painter.base_size))

    # check whether an element interacts with square
    # TODO: this only works if element is a circle!
    def check_interaction(self,element):
        # allow for error
        return l_2(element.pos-self.pos)<(self.size+element.size)/2-1e-11

    # whether x is in the support of the element
    def in_support(self,x):
        return l_2(self.pos-x)<=1/2

    # check whether an element is touching self
    # TODO: this only works if element is a circle!
    def check_touch(self,element):
        # allow for error
        if in_interval(l_2(element.pos-self.pos),(self.size+element.size)/2-1e-11,(self.size+element.size)/2+1e-11):
                return True
        return False

    # find position along a line that comes in contact with the line going through element.pos in direction v
    # TODO: this only works if element is a circle!
    def move_on_line_to_stick(self,element,v):
        # relative position
        x=element.pos-self.pos
        # radius of collision circle
        R=(element.size+self.size)/2

        # smallest root of t^2 v^2+2x.v t+x^2-R^2
        t=(-v.dot(x)-math.sqrt(v.dot(x)*v.dot(x)-v.dot(v)*(x.dot(x)-R*R)))/v.dot(v)

        # return difference to pos
        return v*t

    # move along edge of circle
    # TODO: this only works if element is a circle!
    def move_along(self,delta,element):
        x=element.pos-self.pos+delta
        return x/l_2(x)*(element.size+self.size)/2+self.pos-element.pos

    # for use with lattices
    # list of lattice points covered by square
    def lattice_points(self,lattice):
        out=[]
        dx=math.floor(0.5*self.size/lattice.spacing)
        for i in range(-dx,dx+1):
            for j in range(-dx,dx+1):
                if i*i+j*j<=2*dx*dx:
                    out.append(Point(self.pos.x+i*lattice.spacing,self.pos.y+j*lattice.spacing))
        return out


